# Copyright 2018 DeepMind Technologies Limited. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A simple CSV logger.

Warning: Does not support preemption.
"""

import csv
import os
import time
from typing import TextIO, Union

from absl import logging

from moss.utils import paths
from moss.utils.loggers import base


class CSVLogger(base.Logger):
  """Standard CSV logger.

  The fields are inferred from the first call to write() and any additional
  fields afterwards are ignored.

  TODO(jaslanides): Consider making this stateless/robust to preemption.
  """

  _open = open

  def __init__(
    self,
    directory_or_file: Union[str, TextIO] = '~/moss',
    label: str = '',
    time_delta: float = 0.,
    add_uid: bool = True,
    flush_every: int = 30,
  ):
    """Instantiates the logger.

    Args:
      directory_or_file: Either a directory path as a string, or a file TextIO
        object.
      label: Extra label to add to logger. This is added as a suffix to the
        directory.
      time_delta: Interval in seconds between which writes are dropped to
        throttle throughput.
      add_uid: Whether to add a UID to the file path. See `paths.process_path`
        for details.
      flush_every: Interval (in writes) between flushes.
    """
    if flush_every <= 0:
      raise ValueError(
        f'`flush_every` must be a positive integer (got {flush_every}).'
      )

    self._last_log_time = time.time() - time_delta
    self._time_delta = time_delta
    self._flush_every = flush_every
    self._add_uid = add_uid
    self._writer = None
    self._file_owner = False
    self._file = self._create_file(directory_or_file, label)
    self._writes = 0
    logging.info('Logging to %s', self.file_path)

  def _create_file(
    self,
    directory_or_file: Union[str, TextIO],
    label: str,
  ) -> TextIO:
    """Opens a file if input is a directory or use existing file."""
    if isinstance(directory_or_file, str):
      directory = paths.process_path(
        directory_or_file, label, add_uid=self._add_uid
      )
      file_path = os.path.join(directory, 'logs.csv')
      self._file_owner = True
      return self._open(file_path, mode='a')  # type: ignore

    # TextIO instance.
    file = directory_or_file
    if label:
      logging.info('File, not directory, passed to CSVLogger; label not used.')
    if not file.mode.startswith('a'):
      raise ValueError(
        'File must be open in append mode; instead got '
        f'mode="{file.mode}".'
      )
    return file

  def write(self, data: base.LoggingData) -> None:
    """Writes a `data` into a row of comma-separated values."""
    # Only log if `time_delta` seconds have passed since last logging event.
    now = time.time()

    # TODO(b/192227744): Remove this in favour of filters.TimeFilter.
    elapsed = now - self._last_log_time
    if elapsed < self._time_delta:
      logging.debug(
        'Not due to log for another %.2f seconds, dropping data.',
        self._time_delta - elapsed
      )
      return
    self._last_log_time = now

    # Append row to CSV.
    data = base.to_numpy(data)
    # Use fields from initial `data` to create the header. If extra fields are
    # present in subsequent `data`, we ignore them.
    if not self._writer:
      fields = sorted(data.keys())
      self._writer = csv.DictWriter(
        self._file, fieldnames=fields, extrasaction='ignore'
      )  # type: ignore
      # Write header only if the file is empty.
      if not self._file.tell():
        self._writer.writeheader()  # type: ignore
    self._writer.writerow(data)  # type: ignore

    # Flush every `flush_every` writes.
    if self._writes % self._flush_every == 0:
      self.flush()
    self._writes += 1

  def close(self) -> None:
    """Close."""
    self.flush()
    if self._file_owner:
      self._file.close()

  def flush(self) -> None:
    """Flush."""
    self._file.flush()

  @property
  def file_path(self) -> str:
    """File path."""
    return self._file.name



"""PettingZoo agent."""
from typing import Any, Dict, Tuple

import jax.numpy as jnp

from moss.agent.base import BaseAgent
from moss.core import Buffer, Predictor
from moss.network.keys import AGENT_STATE
from moss.types import LoggingData, Reward, TimeStep
from moss.utils.loggers import Logger


class PettingZooAgent(BaseAgent):
  """PettingZoo agent."""

  def __init__(
    self,
    unroll_length: int,
    buffer: Buffer,
    predictor: Predictor,
    logger: Logger,
    data_format: str = "NHWC"
  ) -> None:
    """Init.

    Args:
      unroll_length: Unroll length.
      buffer: Agent replay buffer.
      predictor: Predictor.
      logger: Logger.
      data_format: PettingZoo image data format, must be `NHWC` or `NCHW`,
        default is `NHWC`.
    """
    super().__init__("PettingZoo", unroll_length, buffer, predictor, logger)
    if data_format not in ["NHWC", "NCHW"]:
      raise ValueError(
        f"data_format must be `NHWC` or `NCHW`, but got `{data_format}`."
      )
    self._data_format = data_format
    self._episode_steps: int = 0
    self._rewards: float = 0

  def _init(self) -> None:
    """Init agent states."""
    self._episode_steps = 0
    self._rewards = 0

  def reset(self) -> LoggingData:
    """Reset agent."""
    metrics = {
      f"{self._name}/episode steps": self._episode_steps,
      f"{self._name}/total rewards": self._rewards
    }
    self._init()
    return metrics

  def step(self, timestep: TimeStep) -> Tuple[Dict, Reward]:
    """Agent step.

    Return:
      input_dict: input dict for network.
        Returns must be serializable Python object to ensure that it can
        exchange data between launchpad's nodes.
    """
    obs = timestep.observation.obs
    info = timestep.observation.info
    if self._data_format == "NCHW":
      obs = jnp.transpose(obs, axes=(2, 1, 0))
    state = {
      "pettingzoo_frame": {
        "frame": jnp.array(obs)
      },
      "player": {
        "id": 0 if info.get("agent_id") == "first_0" else 1
      },
    }
    input_dict = {AGENT_STATE: state}
    reward = timestep.reward * 1.0
    self._episode_steps += 1
    self._rewards += reward
    return input_dict, reward

  def take_action(self, action: Dict[str, Any]) -> Any:
    """Take action."""
    return action["pettingzoo_action"]



"""Base environment."""
import abc
from typing import Any, Dict

from moss.types import Environment, TimeStep

AgentID = Any


class BaseEnv(Environment):
  """Abstract base environments class for moss."""

  @abc.abstractmethod
  def reset(self) -> Dict[AgentID, TimeStep]:
    """Starts a new environment."""

  @abc.abstractmethod
  def step(self, action: Any) -> Dict[AgentID, TimeStep]:
    """Updates the environment."""

  def send(self, action: Any) -> Any:
    """Send action to low-level environment api."""
    raise NotImplementedError

  def recv(self) -> Any:
    """Receive result from low-level environment api."""
    raise NotImplementedError

  def observation_spec(self) -> Any:
    """Defines the observations provided by the environment.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    pass

  def action_spec(self) -> Any:
    """Defines the actions that should be provided to `step`.

    May use a subclass of `specs.Array` that specifies additional properties
    such as min and max bounds on the values.

    Returns:
      An `Array` spec, or a nested dict, list or tuple of `Array` specs.
    """
    pass



"""Revber buffer."""
import threading
from typing import Any, Optional

import jax.numpy as jnp
import reverb
from jax.tree_util import tree_map
from reverb import reverb_types

from moss.core import Buffer
from moss.types import Transition


class ReverbQueue(Buffer):
  """Reverb queue buffer."""

  def __init__(
    self,
    max_size: int,
    signature: Optional[reverb_types.SpecNest] = None,
  ) -> None:
    """Init."""
    self._table = reverb.Table.queue("queue", max_size, signature=signature)
    self._server = reverb.Server([self._table])
    self._client = reverb.Client(f"localhost:{self._server.port}")
    self._mutex = threading.Lock()

  def add(self, data: Any) -> None:
    """Add trajectory data to replay buffer.

    Args:
      data: Trajectory data with shape [T, ...].
    """
    with self._mutex:
      with self._client.trajectory_writer(1) as writer:
        writer.append(
          dict(
            step_type=data.step_type,
            input_dict=data.input_dict,
            action=data.action,
            rnn_state=data.rnn_state,
            reward=data.reward,
            policy_logits=data.policy_logits,
            behaviour_value=data.behaviour_value,
          )
        )
        writer.create_item(
          table=self._table.name,
          priority=1.0,
          trajectory=[
            writer.history["step_type"][0],
            writer.history["input_dict"][0],
            writer.history["action"][0],
            writer.history["rnn_state"][0],
            writer.history["reward"][0],
            writer.history["policy_logits"][0],
            writer.history["behaviour_value"][0],
          ]
        )

  def sample(self, sample_size: int) -> Any:
    """Sample trajectory data from replay buffer.

    Returns:
      Batched trajecotry data with shape [T, B, ...].
    """
    sample = self._client.sample(
      self._table.name, sample_size, emit_timesteps=False
    )
    datas = []
    for data in sample:
      transition = Transition(
        step_type=data.data[0],
        input_dict=data.data[1],
        action=data.data[2],
        rnn_state=data.data[3],
        reward=data.data[4],
        policy_logits=data.data[5],
        behaviour_value=data.data[6],
      )
      datas.append(transition)
    stacked_data = tree_map(lambda *x: jnp.stack(x, axis=1), *datas)
    return stacked_data
